﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Parameters;
using HeuristicLab.Encodings.PermutationEncoding;
using System.ComponentModel;

namespace HeuristicLab.Problems.Scheduling {
  [Item("Schedule Class", "Represents the general solution for scheduling problems.")]
  [StorableClass]
  public class Schedule : NamedItem{
    [StorableConstructor]
    protected Schedule(bool deserializing) : base(deserializing) { }
    protected Schedule(Schedule original, Cloner cloner)
      : base(original, cloner) {
        this.Resources = cloner.Clone (original.Resources);
        this.Quality = cloner.Clone(original.Quality);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new Schedule(this, cloner);
    }

    [Storable]
    private ItemList<Resource> resources;
    public ItemList<Resource> Resources {
      get { return resources; }
      set {
        if (resources != null) DeregisterResourcesEvents();
        resources = value;
        if (resources != null) RegisterResourcesEvents();
        OnResourcesChanged();
      }
    }

    [Storable]
    private DoubleValue quality;
    public DoubleValue Quality {
      get { return quality; }
      set {
        if (quality != value) {
          if (quality != null) DeregisterQualityEvents();
          quality = value;
          if (quality != null) RegisterQualityEvents();
          OnQualityChanged();
        }
      }
    }

    public double Makespan {
      get {
        double quality = 0;
        foreach (Resource r in Resources) {
          if (r.TotalDuration.Value > quality) {
            quality = r.TotalDuration.Value;
          }
        }
        return quality;
      }
    }

    public Schedule(IntValue nrOfResources) {
      Resources = new ItemList<Resource>();
      Quality = new DoubleValue();
      for (int i = 0; i < nrOfResources.Value; i++) {
        Resources.Add(new Resource(new IntValue (i)));
      }
    }

    #region Events
    public event EventHandler QualityChanged;
    private void OnQualityChanged() {
      var changed = QualityChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }
    private void RegisterQualityEvents() {
      Quality.ValueChanged += new EventHandler(Quality_ValueChanged);
    }
    private void DeregisterQualityEvents() {
      Quality.ValueChanged -= new EventHandler(Quality_ValueChanged);
    }
    private void Quality_ValueChanged(object sender, EventArgs e) {
      OnQualityChanged();
    }

    public event EventHandler ResourcesChanged;
    private void OnResourcesChanged() {
      var changed = ResourcesChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }
    private void RegisterResourcesEvents() {
      Resources.PropertyChanged += new System.ComponentModel.PropertyChangedEventHandler(Resources_PropertyChanged);
    }
    private void DeregisterResourcesEvents() {
      Resources.PropertyChanged -= new System.ComponentModel.PropertyChangedEventHandler(Resources_PropertyChanged);
    }
    private void Resources_PropertyChanged(object sender, EventArgs e) {
      OnResourcesChanged();
    }
    #endregion

    public void AddTask(JSSPTask task, JSSPJob job) {
      task.StartTime = new DoubleValue (ComputeEarliestStartTime(task, job));
      Resource affectedResource = resources[task.ResourceNr.Value];
      int i = 0;
      while (i < affectedResource.Tasks.Count && affectedResource.Tasks[i].StartTime.Value < task.StartTime.Value)
        i++;

      if (i >= affectedResource.Tasks.Count)
        affectedResource.Tasks.Add(task);
      else
        affectedResource.Tasks.Insert(i, task);
    }
    public double ComputeEarliestStartTime(JSSPTask t, JSSPJob job) {
      Resource affectedResource = resources[t.ResourceNr.Value];
      double lastMachineEndTime = affectedResource.TotalDuration.Value;
      double previousJobTaskEndTime = 0;
      JSSPTask previousTask = job.GetPreviousTask(t);
      if (previousTask != null)
        previousJobTaskEndTime = previousTask.EndTime.Value;

      return Math.Max(previousJobTaskEndTime, lastMachineEndTime);
    }
    public double ComputeEarliestCompletionTime(JSSPTask t, JSSPJob job) {
      Resource affectedResource = resources[t.ResourceNr.Value];
      return ComputeEarliestStartTime(t, job) + t.Duration.Value;
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append("[ ");
      foreach (Resource r in Resources) {
        sb.Append(r.ToString () + " \n");
      }
      sb.Append("]");
      return sb.ToString();
    }
  }
}
